/**
 * @brief Adaptive logarithmic tone mapping
 * 
 * Adaptive logarithmic mapping for displaying high contrast
 * scenes. 
 * F. Drago, K. Myszkowski, T. Annen, and N. Chiba. In Eurographics 2003.
 *
 * This file is a part of PFSTMO package.
 * ---------------------------------------------------------------------- 
 * Copyright (C) 2003,2004 Grzegorz Krawczyk
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ---------------------------------------------------------------------- 
 * 
 * @author Grzegorz Krawczyk, <krawczyk@mpi-sb.mpg.de>
 *
 * $Id: pfstmo_drago03.cpp,v 1.1 2004/09/22 10:00:23 krawczyk Exp $
 */

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <pfs.h>
#include <QFile>
#include <tmo_drago03.h>

void calculateLuminance( pfs::Array2D* Y, float& avLum, float& maxLum );

#ifdef _WIN32
QByteArray pfstmo_drago03(const char* input, const char* output, float _biasValue) {
	FILE *inputfp=fopen(input,"rb");
        FILE *fpout=fopen(output,"wb");
#else
QByteArray pfstmo_drago03(QByteArray inqba, float _biasValue) {
	assert(inqba.size()!=0);
	FILE *inputfp=fmemopen(inqba.data(), (size_t)(inqba.size()), "r");
	//prepare "fake" FILE* where pfs library will write the output hdr to, in pfs stream format.
	char *pfshdr;
	size_t size;
	FILE * fpout=open_memstream (&pfshdr, &size);
#endif
	pfs::DOMIO pfsio;

	//--- default tone mapping parameters;
	float biasValue = _biasValue;

	
	pfs::Frame *frame = pfsio.readFrame( inputfp );
	assert(frame != NULL);

	pfs::Channel *X, *Y, *Z;
	frame->getXYZChannels( X, Y, Z );

	float maxLum,avLum;
	calculateLuminance( Y, avLum, maxLum );
	
	int w = Y->getCols();
	int h = Y->getRows();

	pfs::Array2D* L = new pfs::Array2DImpl(w,h);
	tmo_drago03(Y, L, maxLum, avLum, biasValue);

	for( int x=0 ; x<w ; x++ )
		for( int y=0 ; y<h ; y++ ) 
		{
			float scale = (*L)(x,y) / (*Y)(x,y);
			(*Y)(x,y) *= scale;
			(*X)(x,y) *= scale;
			(*Z)(x,y) *= scale;
		}

	delete L;

	pfsio.writeFrame( frame, fpout );
	pfsio.freeFrame( frame ); // would this delete the previous buffer?

	fclose(fpout);
	fclose(inputfp);
	QByteArray toreturn;
#ifndef _WIN32
	toreturn=QByteArray::fromRawData(pfshdr,size); //shallow copy
// 	free(pfshdr); //since QByteArray::fromRawData is shallow copy, I cannot free it.
#else
	QFile f(output);
	f.open(QIODevice::ReadOnly);
	toreturn=f.readAll();
#endif
	return toreturn;
}

void calculateLuminance( pfs::Array2D* Y, float& avLum, float& maxLum )
{
  avLum = 0.0f;
  maxLum = 0.0f;

  int size = Y->getCols() * Y->getRows();

  for( int i=0 ; i<size; i++ )
  {
    avLum += log( (*Y)(i) + 1e-4 );
    maxLum = ( (*Y)(i) > maxLum ) ? (*Y)(i) : maxLum ;
  }
  avLum =exp( avLum/ size);
}
