/**
 * @file pfstmo_durand02.cpp
 * @brief Tone map XYZ channels using Durand02 model
 *
 * Fast Bilateral Filtering for the Display of High-Dynamic-Range Images.
 * F. Durand and J. Dorsey.
 * In ACM Transactions on Graphics, 2002.
 *
 * 
 * This file is a part of PFSTMO package.
 * ---------------------------------------------------------------------- 
 * Copyright (C) 2003,2004 Grzegorz Krawczyk
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ---------------------------------------------------------------------- 
 * 
 * @author Grzegorz Krawczyk, <krawczyk@mpi-sb.mpg.de>
 *
 * $Id: pfstmo_durand02.cpp,v 1.5 2005/12/15 15:53:37 krawczyk Exp $
 */
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <pfs.h>
#include <tmo_durand02.h>
#include <QFile>

using namespace std;

#ifdef _WIN32
QByteArray pfstmo_durand02(const char* input, const char* output, float _sigma_s, float _sigma_r, float _baseContrast) {
	FILE *inputfp=fopen(input,"rb");
        FILE *fpout=fopen(output,"wb");
#else
QByteArray pfstmo_durand02(QByteArray inqba, float _sigma_s, float _sigma_r, float _baseContrast) {
	assert(inqba.size()!=0);
	FILE *inputfp=fmemopen(inqba.data(), (size_t)(inqba.size()), "r");
	//prepare "fake" FILE* where pfs library will write the output hdr to, in pfs stream format.
	char *pfshdr;
	size_t size;
	FILE * fpout=open_memstream (&pfshdr, &size);
#endif
  pfs::DOMIO pfsio;

// #ifdef HAVE_FFTW
//   float sigma_s = 40.0f;
// #else
//   float sigma_s = 8.0f;
// #endif
//   float sigma_r = 0.4f;
//   float baseContrast = 5.0f;
	float sigma_s=_sigma_s;
	float sigma_r=_sigma_r;
	float baseContrast=_baseContrast;
	int downsample=1;

#ifdef HAVE_FFTW
//  qDebug("fast bilateral filtering (fftw3)");
#else
//  qDebug("conventional bilateral filtering");
#endif

	pfs::Frame *frame = pfsio.readFrame( inputfp );
	assert( frame != NULL );
	
	pfs::Channel *X, *Y, *Z;
	frame->getXYZChannels( X, Y, Z );
	
	assert( Y!=NULL || X!=NULL || Z!=NULL);
	
	pfs::transformColorSpace( pfs::CS_XYZ, X, Y, Z, pfs::CS_RGB, X, Y, Z );
	tmo_durand02( X, Y, Z, sigma_s, sigma_r, baseContrast, downsample );
	pfs::transformColorSpace( pfs::CS_RGB, X, Y, Z, pfs::CS_XYZ, X, Y, Z );
	
	pfsio.writeFrame( frame, fpout );
	pfsio.freeFrame( frame );

	fclose(inputfp);
	fclose(fpout);
	QByteArray toreturn;
#ifndef _WIN32
	toreturn=QByteArray::fromRawData(pfshdr,size); //shallow copy
// 	free(pfshdr); //since QByteArray::fromRawData is shallow copy, I cannot free it.
#else
	QFile f(output);
	f.open(QIODevice::ReadOnly);
	toreturn=f.readAll();
#endif
	return toreturn;
}
