#include <pfs.h>
#include <math.h>
#include <stdio.h>
#include "createhdr.h"


#ifdef _WIN32
void       createHDR(QList<QImage*> &list, const float * arrayofexptime, config_triple* chosen_config, bool antighosting, int iterations, const char * outfname) {
    FILE *fpout=fopen(outfname,"wb");
#else
QByteArray createHDR(QList<QImage*> &list, const float * arrayofexptime, config_triple* chosen_config, bool antighosting, int iterations){
    //prepare "fake" FILE* where pfs library will write the output hdr to, in pfs stream format.
    char *pfshdr;
    size_t size;
    FILE * fpout=open_memstream(&pfshdr, &size);
#endif

    int opt_bpp=8;
    float opt_gauss = 8.0f;
    int opt_minresponse=0;
    int opt_maxresponse=255;

    pfs::DOMIO pfsio;

    TWeight opt_weight = chosen_config->weights;
    TResponse opt_response = chosen_config->response_curve;
    TModel opt_model = chosen_config->model;
    const int width=(list.at(0))->width();
    const int height=(list.at(0))->height();
    long sp=0;

    if (antighosting) {
    	qDebug("ag, setting model debevec and weights plateau");
    	opt_weight=PLATEAU;
    	opt_model=DEBEVEC;
    }

    //PRINT WEIGHTS PREFERENCES
    switch( opt_weight )
    {
    case TRIANGULAR:
	qDebug("using triangular-shaped weights");
	break;
    case GAUSSIAN:
	qDebug( "using gaussian-shaped weights");
	break;
    case PLATEAU:
	qDebug( "using plateau-shaped weights");
	break;
    default:
	throw pfs::Exception("weights not recognized");
    }

//PRINT RESPONSE CURVES PREFERENCES
    switch( opt_response )
    {
    case FROM_FILE:
	qDebug("loading response curve from file");
	break;
    case LINEAR:
	qDebug( "initial response: linear" );
	break;
    case GAMMA:
	qDebug( "initial response: gamma");
	break;
    case LOG10:
	qDebug("initial response: logarithmic");
	break;
    case FROM_ROBERTSON:
	qDebug( "response curve to be found from image set");
	break;
    default:
	throw pfs::Exception("undefined standard response");
	break;
  }

//PRINT HDR-GENERATION MODEL PREFERENCES
    switch( opt_model )
    {
    case ROBERTSON:
	qDebug("using robertson model");
	break;
    case DEBEVEC:
	qDebug("using debevec model");
	break;
    default:
	throw pfs::Exception("hdr generation method not set or not supported");
    }
    

    int M = (int) powf(2.0f,opt_bpp);
    // weighting function representing confidence in accuracy of acquisition    
    float* w = new float[M];
    //1) fill in w based on weights preferences
    switch( opt_weight )
    {
    case TRIANGULAR:
	weights_triangle(w, M);
	break;
    case GAUSSIAN:
	weightsGauss( w, M, opt_minresponse, opt_maxresponse, opt_gauss );
	break;
    case PLATEAU:
	exposure_weights_icip06(w, M);
	break;
    }
    // create channels for output
    pfs::Frame *frameout = pfsio.createFrame( width, height );
    pfs::Channel *Xj, *Yj, *Zj;
    frameout->createXYZChannels( Xj, Yj, Zj );
    // camera response functions for each channel
    float* Ir = new float[M];
    float* Ig = new float[M];
    float* Ib = new float[M];
    //2) response curves, either predefined (log,gamma,lin,from_file) or calibration from the set of images.
    switch( opt_response )
    {
    case FROM_FILE:
	{
	    FILE * respfile=fopen(chosen_config->CurveFilename.toAscii().constData(),"r");
	    // read camera response from file
	    bool loadR_ok = responseLoad(respfile, Ir, M);
	    bool loadG_ok = responseLoad(respfile, Ig, M);
	    bool loadB_ok = responseLoad(respfile, Ib, M);
	    fclose(respfile);
	    if( !loadR_ok || !loadG_ok || !loadB_ok )
		throw pfs::Exception( "could not load response curve from file" );
	}
	break;
    case LINEAR:
	responseLinear( Ir, M );
	responseLinear( Ig, M );
	responseLinear( Ib, M );
	break;
    case GAMMA:
	responseGamma( Ir, M );
	responseGamma( Ig, M );
	responseGamma( Ib, M );
	break;
    case LOG10:
	responseLog10( Ir, M );
	responseLog10( Ig, M );
	responseLog10( Ib, M );
	break;
    case FROM_ROBERTSON:
	responseLinear( Ir, M );
	responseLinear( Ig, M );
	responseLinear( Ib, M );
	//call robertson02_getResponse method which computes both the Ir,Ig,Ib and the output HDR (i.e. its channels Xj,Yj,Zj).
//	qDebug( "recovering R channel...");
	sp   = robertson02_getResponse( Xj, list, arrayofexptime, Ir, w, M, 1);
//	qDebug("recovering G channel...");
	sp += robertson02_getResponse( Yj, list, arrayofexptime, Ig, w, M, 2);
//	qDebug("recovering B channel...");
	sp += robertson02_getResponse( Zj, list, arrayofexptime, Ib, w, M, 3);
	sp /= 3;
	break;
    }
    
    //3) apply model to generate hdr. If model is robertson and user preference was to compute response curve from image dataset using robertson algorithm, i.e. to calibrate, we are done, the robertson02_getResponse function has already computed the HDR (i.e. its channels).
    switch( opt_model )
    {
    case ROBERTSON:
	if (opt_response==FROM_ROBERTSON)
	    break;
	else {
	    //apply robertson model
//	    qDebug("applying robertson model to R channel...");
	    sp =  robertson02_applyResponse( Xj, list, arrayofexptime, Ir, w, M, 1);
//	    qDebug("applying robertson model to G channel...");
	    sp += robertson02_applyResponse( Yj, list, arrayofexptime, Ig, w, M, 2);
//	    qDebug("applying robertson model to B channel...");
	    sp += robertson02_applyResponse( Zj, list, arrayofexptime, Ib, w, M, 3);
	}
	break;
    case DEBEVEC:
	//apply devebec model
	if (antighosting) {
		Array2DList Ptemp,P;
		icip06_applyResponse(list, arrayofexptime, Xj, Yj, Zj, Ir, Ig, Ib, w, iterations, Ptemp, P);
		}
	else
		devebec_applyResponse(  Xj, Yj, Zj, list, arrayofexptime, Ir, Ig, Ib, w, M);
	break;
    }

    // output PFS stream with calibrated response
    pfs::transformColorSpace( pfs::CS_RGB, Xj, Yj, Zj, pfs::CS_XYZ, Xj, Yj, Zj );
    pfsio.writeFrame( frameout, fpout );
//     FILE * fp2=fopen("out.pfs","w");
//     pfsio.writeFrame( frameout, fp2 );
//     (list.at(0))->save("a.jpg","JPEG");

    delete[] w;
    delete[] Ir;
    delete[] Ig;
    delete[] Ib;

    fclose(fpout);
//    qDebug("written file: %s",outfname);
#ifndef _WIN32
    QByteArray toreturn;
    toreturn=QByteArray::fromRawData(pfshdr,size); //shallow copy
//     free(pfshdr); //since fromRawData is shallow copy, I cannot free it.
//    qDebug("size=%d",toreturn.size());
    return toreturn;
#endif
//     qDebug(QString("size=%1").arg(toreturn.size()).toAscii().constData());
}
