#include "imagehdrviewer.h"
#include <stdio.h>
#include <math.h>
#include <QApplication>
#include <QLabel>
#include <QFileInfo>

inline float clamp( float val, float min, float max )
{
    if( val < min ) return min;
    if( val > max ) return max;
    return val;
}


inline int clamp( int val, int min, int max )
{
    if( val < min ) return min;
    if( val > max ) return max;
    return val;
}


static float getInverseMapping( LumMappingMethod mappingMethod,
				float v, float minValue, float maxValue ) {
    switch( mappingMethod ) {
    case MAP_GAMMA1_4:
	return powf( v, 1.4 )*(maxValue-minValue) + minValue;
    case MAP_GAMMA1_8:
	return powf( v, 1.8 )*(maxValue-minValue) + minValue;
    case MAP_GAMMA2_2:
	return powf( v, 2.2 )*(maxValue-minValue) + minValue;
    case MAP_GAMMA2_6:
	return powf( v, 2.6 )*(maxValue-minValue) + minValue;
    case MAP_LINEAR:
	return v*(maxValue-minValue) + minValue;
    case MAP_LOGARITHMIC:
	return powf( 10, v * (log10f(maxValue) - log10f(minValue)) + log10f( minValue ) );
    default:
	assert(0);
	return 0;
    }
}


inline int binarySearchPixels( float x, const float *lut, const int lutSize ) {
    int l = 0, r = lutSize;
    while( true ) {
	int m = (l+r)/2;
	if( m == l ) break;
	if( x < lut[m] )
	    r = m;
	else
	    l = m;
    }
    return l;
}


static void mapFrameToImage( pfs::Array2D *R, pfs::Array2D *G, pfs::Array2D *B, 
			     QImage *img,
			     float minValue, float maxValue, LumMappingMethod mappingMethod) {

    int rows = R->getRows();
    int cols = R->getCols();
    int index = 0;
    float imgRow = 0, imgCol;
    float lutPixFloor[257*2];
    QRgb lutPixel[257*2];
    int lutSize=258;

    for( int p = 1; p <= 257; p++ ) {
	float p_left = ((float)p - 1.f)/255.f; // Should be -1.5f, but we don't want negative nums
	lutPixFloor[p] = getInverseMapping( mappingMethod, p_left, minValue, maxValue );
    }
    lutPixel[0] = QColor( 0, 0, 0 ).rgb();
    lutPixel[257] = QColor( 0, 0, 0 ).rgb();

    for( int r = 0; r < rows; r++, imgRow++ ) {
	QRgb* line = (QRgb*)img->scanLine( (int)imgRow );
	imgCol = 0;
	for( int c = 0; c < cols; c++, index++, imgCol++ ) {
	    QRgb pixel;
	    // Color channels
	    int pr, pg, pb;
	    pr = binarySearchPixels( (*R)(index), lutPixFloor, lutSize );
	    pg = binarySearchPixels( (*G)(index), lutPixFloor, lutSize );
	    pb = binarySearchPixels( (*B)(index), lutPixFloor, lutSize );

	    // Clipping
	    pixel = QColor( clamp( pr-1, 0, 255 ),
			    clamp( pg-1, 0, 255 ),
			    clamp( pb-1, 0, 255 ) ).rgb();

	    if( !finite( (*R)(index) ) || !finite( (*G)(index) ) || !finite( (*B)(index) ) )
	    {   // x is NaN or Inf 
		pixel = QColor( 255, 0, 0 ).rgb();
	    }

	    if( (*R)(index)<0 || (*G)(index)<0 || (*B)(index)<0 ) {   // x is negative
	    pixel = QColor( 0, 0, 0 ).rgb();
	  }

	    line[(int)imgCol] = pixel;
	}
    }
}

static void transformImageToWorkArea( pfs::Array2D **workArea,
				      pfs::Array2D *X, pfs::Array2D *Y, pfs::Array2D *Z ) {

    for( int c = 0; c < 3; c++ ) {
	if( workArea[c] == NULL ) {
	    workArea[c] = new pfs::Array2DImpl( X->getCols(), X->getRows() );
	}
	// workArea has different colums/rows when rotating
	if ( workArea[c]->getCols() != X->getCols()) {
	    delete workArea[c] ;
	    workArea[c]=new pfs::Array2DImpl( X->getCols(), X->getRows() );
	}
    }
    //Copy & tranform image to work area
    copyArray( X, workArea[0] );
    copyArray( Y, workArea[1] );
    copyArray( Z, workArea[2] );

    pfs::transformColorSpace( pfs::CS_XYZ, workArea[0], workArea[1], workArea[2],
			         pfs::CS_RGB, workArea[0], workArea[1], workArea[2] );
}


//==================================================================================
//**********************************************************************************
//==================================================================================

ImageMDIwindow::ImageMDIwindow( QWidget *parent) : QMainWindow(parent), pfsFrame(NULL), mappingMethod( MAP_GAMMA2_2 ), minValue( 1.0f ), maxValue( 1.0f ), fittingwin(false) {

    setAttribute(Qt::WA_DeleteOnClose);
//     qDebug(QString("in imagemdi(), size=%1").arg(qba.size()).toAscii());
    image=NULL;
    workArea[0] = workArea[1] = workArea[2] = NULL;

    imageLabel = new QLabel;
    imageLabel->setBackgroundRole(QPalette::Base);
    imageLabel->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Ignored);
    imageLabel->setScaledContents(true); ///true in qt4 example

    scrollArea = new QScrollArea;
    scrollArea->setBackgroundRole(QPalette::Dark);
    scrollArea->setWidget(imageLabel);
//     scrollArea->setWidgetResizable(false); //true if you want to "fit the window" in qt4 example
    setCentralWidget(scrollArea);

    toolBar = this->addToolBar("Viewing Settings Toolbar");
    QLabel *mappingMethodLabel = new QLabel( "&Mapping:", toolBar );
    mappingMethodLabel->setAlignment( Qt::AlignRight | Qt::AlignVCenter );
    mappingMethodCB = new QComboBox( toolBar );
    toolBar->addWidget(mappingMethodLabel);
    toolBar->addWidget(mappingMethodCB);
    toolBar->addSeparator();
    mappingMethodLabel->setBuddy( mappingMethodCB );
    QStringList methods;
    methods+="Linear";
    methods+="Gamma 1.4";
    methods+="Gamma 1.8";
    methods+="Gamma 2.2";
    methods+="Gamma 2.6";
    methods+="Logarithmic";
    mappingMethodCB->addItems(methods);
    mappingMethodCB->setCurrentIndex( 3 );
    connect( mappingMethodCB, SIGNAL( activated( int ) ), this, SLOT( setLumMappingMethod(int) ) );

    QLabel *histlabel = new QLabel( "Histogram:", toolBar );
    lumRange = new LuminanceRangeWidget( toolBar );
    toolBar->addWidget(histlabel);
    toolBar->addWidget(lumRange);
    toolBar->addSeparator();
    connect( lumRange, SIGNAL( updateRangeWindow() ), this, SLOT( updateRangeWindow() ) );
}

void ImageMDIwindow::updateRangeWindow() {
 setRangeWindow( pow( 10, lumRange->getRangeWindowMin() ), pow( 10, lumRange->getRangeWindowMax() ) );
}

#ifdef _WIN32
void ImageMDIwindow::updateHDR(const char *infname) {
    FILE *fp=fopen(infname,"rb");
#else
void ImageMDIwindow::updateHDR(QByteArray qba) {
    assert(qba.size()!=0);
    FILE *fp=fmemopen(qba.data(), (size_t)(qba.size()), "r");
    hdrpfsstream.clear(); //previous hdr pfs stream not valid anymore
    hdrpfsstream=qba;
#endif
    pfs::DOMIO pfsCtx;
    if (pfsFrame!=NULL)
       delete pfsFrame;
//    if (fp==NULL) qDebug("FP IS NULL!!!!!!");
    pfsFrame = pfsCtx.readFrame( fp );
    assert(pfsFrame != NULL);
    lumRange->setHistogramImage(getPrimaryChannel());
    lumRange->fitToDynamicRange();
    updateImage();
    fclose(fp);
    //zoom is original size, 100%
    scaleFactor = 1.0;
    //make the label use the image dimensions
    imageLabel->adjustSize();
    update();
    imageLabel->update();
    repaint();
    imageLabel->repaint();
}

void ImageMDIwindow::updateImage() {
	assert( pfsFrame != NULL );
	
	QApplication::setOverrideCursor( QCursor(Qt::WaitCursor) );
	
	pfs::Channel *X, *Y, *Z;
// 	qDebug("before getxyzchans");
	pfsFrame->getXYZChannels( X, Y, Z );
	assert(X!=NULL || Y!=NULL || Z!=NULL);
	
	//workarea is the RGB version of XYZ; workarea needed in updateMapping(...) called by mapFrameToImage(...)
	transformImageToWorkArea( workArea, X, Y, Z );
// 	qDebug("after transformImageToWorkArea");
	
	int zoomedWidth = workArea[0]->getCols();
	int zoomedHeight = workArea[0]->getRows();
	if( image != NULL )
		delete image;
	image = new QImage(zoomedWidth, zoomedHeight, QImage::Format_RGB32 );
	assert( image != NULL );
	
	assert( workArea[0] != NULL && workArea[1] != NULL && workArea[2] != NULL );
	mapFrameToImage( workArea[0], workArea[1], workArea[2], image,
			minValue, maxValue, mappingMethod );
	
	//assign the mapped image to the label
	imageLabel->setPixmap(QPixmap::fromImage(*image));
	QApplication::restoreOverrideCursor();
}

void ImageMDIwindow::setRangeWindow( float min, float max ) {
	minValue = min;
	maxValue = max;
	updateImage();
}
void ImageMDIwindow::setLumMappingMethod( int method ) {
	mappingMethodCB->setCurrentIndex( method );
	mappingMethod = (LumMappingMethod)method;
	updateImage();
}
void ImageMDIwindow::zoomIn() {
// 	scrollArea->setWidgetResizable(false);
	scaleImage(1.25);
}
void ImageMDIwindow::zoomOut() {
// 	scrollArea->setWidgetResizable(false);
	scaleImage(0.8);
}
void ImageMDIwindow::fitToWindow(bool checked) {
	scrollArea->setWidgetResizable(checked);
	fittingwin=checked;
	if (!checked)
		// restore to the previous zoom factor
		scaleImage(1);
}
void ImageMDIwindow::normalSize() {
// 	scrollArea->setWidgetResizable(false);
	imageLabel->adjustSize();
	scaleFactor = 1.0;
}
void ImageMDIwindow::scaleImage(double factor) {
	scaleFactor *= factor;
	imageLabel->resize(scaleFactor * imageLabel->pixmap()->size());
	adjustScrollBar(scrollArea->horizontalScrollBar(), factor);
	adjustScrollBar(scrollArea->verticalScrollBar(), factor);
}
void ImageMDIwindow::adjustScrollBar(QScrollBar *scrollBar, double factor) {
	scrollBar->setValue(int(factor * scrollBar->value() + ((factor - 1) * scrollBar->pageStep()/2)));
}
double ImageMDIwindow::getScaleFactor() {
	return scaleFactor;
}
bool ImageMDIwindow::getFittingWin(){
	return fittingwin;
}

const pfs::Array2D *ImageMDIwindow::getPrimaryChannel() {
    assert( pfsFrame != NULL );
    pfs::Channel *X, *Y, *Z;
    pfsFrame->getXYZChannels( X, Y, Z );
    return Y;
}

ImageMDIwindow::~ImageMDIwindow() {
	delete workArea[0];
	delete workArea[1];
	delete workArea[2];
	delete imageLabel;
	delete scrollArea;
	delete image;
	if (pfsFrame!=NULL)
		delete pfsFrame;
#ifndef _WIN32
	hdrpfsstream.clear();
#endif
}

#ifndef _WIN32
QByteArray ImageMDIwindow::getHdrPfsStream() {
    return hdrpfsstream;
}
#endif
