#include "maingui_impl.h"
#include "imagehdrviewer.h"
#include "tonemappingdialog_impl.h"
#include "../generated_uic/ui_help_about.h"
#include <QFileDialog>
#include <QDir>
#include <QFileInfo>
#include <QMessageBox>
#include "config.h"

#ifdef _WIN32
// void readEXRfile  (const char * filename, const char *outfilename);
void readRGBEfile (const char * filename,const char* outfilename);
void rotateFrame( const char* input, bool clock_wise );
void writeRGBEfile (const char* infilename, const char* outfilename);
// void writeEXRfile  (const char* infilename, const char* outfilename);
#else
QByteArray readEXRfile  (const char * filename);
QByteArray readRGBEfile (const char * filename);
QByteArray rotateFrame( QByteArray inqba, bool clock_wise );
void writeRGBEfile (QByteArray inqba,const char* outfilename);
void writeEXRfile  (QByteArray inqba,const char* outfilename);
#endif

enum { MAX_RECENTFILES = 9 };

static const QString helptext = 
"<p align=\"center\"><b><font size=\"+4\"><u>Q t p f s g u i</u></b></p>\
<p align=\"center\">V"QTPFSGUIVERSION "(Qt4)</p>\
<hr> <h1>Overview and Usage</h1> \
<p>This is a GUI interface to a set of commandline utilities that operate on HDR images.</p> \
<p>The <b>Open</b>&#160;<img src=\":/new/prefix1/images/fileopen.png\">,&#160; <b>New</b>   <img src=\":/new/prefix1/images/newfile.png\"> &#160;and <b>Save as...</b><img src=\":/new/prefix1/images/filesave.png\">&#160; buttons in the main window operate on HDR images: the \"<b>New</b>\" button pops up a wizard. <br>\
Throughout the guided procedure you will be asked about the location of your source images, and the configuration you want to use to create the HDR.</p> \
<h2>The HDR configurations in the wizard</h2>\
<p>Usually one of the 6 default configurations leads to very good results, but if you want you can enable the \"<b>custom configuration</b>\" checkbox that will prompt you with all the possibilities. <br>\
Let me know if you find one combination (that is not already in the default ones) that creates good results for you.</p>\
<br><hr>\
<h1>Tonemapping the HDR into a LDR</h1>\
<p>Once you have a window with a HDR in it you can tonemap it via the <b>Tonemap</b>  <img src=\":/new/prefix1/images/tonemap.png\"> button, that creates a new dialog where you can fiddle with the variours <b><i>tone mapping operators</i></b> (<b>TMO</b>s) and related settings.\
It is usually a good idea to work on a smaller scale image while trying out the TMOs, because they can be very cpu-intensive.<br>\
Once you are satisfied with the results you can click on the Save button:\
<br><img src=\":/new/prefix1/images/filesaveas.png\"><br>\
to save the resulting <b><i>low dynamic range</i></b> (<b>LDR</b>) image.</p>\
<hr><h1>Contact</h1>\
<p>If you use this program, please, please let me know:</p>\
<ul>\
<li>if you find any bug.\
<li>if you find any usability issue.\
</ul>\
<p>Even if you just use this program and find it useful, it would be great if you would let me know.</p>\
Giuseppe Rota<br>\
grota at users.sourceforge.net";


MainGui::MainGui(QWidget *p) : QMainWindow(p), settings("Qtpfsgui", "Qtpfsgui")
{
	setupUi(this);
	connect(this->fileExitAction, SIGNAL(activated()), this, SLOT(close()));
	workspace = new QWorkspace(this);
	workspace->setScrollBarsEnabled( TRUE );
	setCentralWidget(workspace);

	RecentDirHDRSetting=settings.value("RecentDirHDRSetting",QDir::currentPath()).toString();

	setWindowTitle("Qtpfsgui v"QTPFSGUIVERSION);
	connect(workspace, SIGNAL(windowActivated(QWidget*)), this, SLOT(updatecurrentMDIwindow(QWidget *)) );
	connect(fileNewAction, SIGNAL(activated()), this, SLOT(fileNewViaWizard()));
	connect(fileOpenAction, SIGNAL(activated()), this, SLOT(fileOpen()));
	connect(fileSaveAsAction, SIGNAL(activated()), this, SLOT(fileSaveAs()));
	connect(TonemapAction, SIGNAL(activated()), this, SLOT(tonemap_requested()));
	connect(rotateccw, SIGNAL(activated()), this, SLOT(rotateccw_requested()));
	connect(rotatecw, SIGNAL(activated()), this, SLOT(rotatecw_requested()));
	connect(Low_dynamic_range,SIGNAL(activated()),this,SLOT(current_mdiwindow_ldr_exposure()));
	connect(Fit_to_dynamic_range,SIGNAL(activated()),this,SLOT(current_mdiwindow_fit_exposure()));
	connect(Shrink_dynamic_range,SIGNAL(activated()),this,SLOT(current_mdiwindow_shrink_exposure()));
	connect(Extend_dynamic_range,SIGNAL(activated()),this,SLOT(current_mdiwindow_extend_exposure()));
	connect(Decrease_exposure,SIGNAL(activated()),this,SLOT(current_mdiwindow_decrease_exposure()));
	connect(Increase_exposure,SIGNAL(activated()),this,SLOT(current_mdiwindow_increase_exposure()));
	connect(zoomInAct,SIGNAL(activated()),this,SLOT(current_mdiwindow_zoomin()));
	connect(zoomOutAct,SIGNAL(activated()),this,SLOT(current_mdiwindow_zoomout()));
	connect(fitToWindowAct,SIGNAL(toggled(bool)),this,SLOT(current_mdiwindow_fit_to_win(bool)));
	connect(normalSizeAct,SIGNAL(activated()),this,SLOT(current_mdiwindow_original_size()));
// 	connect(menuView,SIGNAL(aboutToShow()),this,SLOT(viewMenuAboutToShow()));
	connect(helpAboutAction,SIGNAL(activated()),this,SLOT(helpAbout()));

        for (int i = 0; i < MaxRecentFiles; ++i) {
            recentFileActs[i] = new QAction(this);
            recentFileActs[i]->setVisible(false);
            connect(recentFileActs[i], SIGNAL(triggered()),
                    this, SLOT(openRecentFile()));
        }
        separatorRecentFiles = menuFile->addSeparator();
	for (int i = 0; i < MaxRecentFiles; ++i)
		menuFile->addAction(recentFileActs[i]);
	updateRecentFileActions();

	this->showMaximized();
	statusBar()->showMessage("Ready.... Now open an HDR or create one!",17000);
}

void MainGui::fileNewViaWizard()
{
	wizard=new HdrWizardForm ();
	if (wizard->exec() == QDialog::Accepted) {
#ifdef _WIN32
		ImageMDIwindow *newmdi=new ImageMDIwindow( this/*,  QByteArray()*/);
		newmdi->updateHDR("original.pfs");
#else
		ImageMDIwindow *newmdi=new ImageMDIwindow( this/*,  wizard->getPfsOutputHDR()*/);
		newmdi->updateHDR(wizard->getPfsOutputHDR());
#endif
		workspace->addWindow(newmdi);
		newmdi->toolBar->addAction(TonemapAction);
		newmdi->setWindowTitle(wizard->getCaptionTEXT());
		newmdi->show();
	}
	if (wizard != NULL)
		delete wizard;
}

void MainGui::fileOpen()
{	
	QString filetypes;
#ifndef _WIN32
	filetypes += "OpenEXR (*.exr);;";
#endif
	filetypes += "Radiance RGBE (*.hdr *.HDR *.pic *.PIC);;";
	filetypes += "PFS stream (*.pfs *.PFS)";
//     filetypes += "TIFF (*.tif *.tiff *.TIFF *.TIF);;";
//     filetypes += "PFM (*.pfm *.PFM);;";
//     filetypes += "Canon 350D RAW (*.cr2 *.CR2);;";
//     filetypes += "Canon EOS RAW (*.CRW *.crw);;";
//     filetypes += "ALL (*)";
	QString opened = QFileDialog::getOpenFileName(
			this,
			"Choose a HDR file to OPEN...",
			RecentDirHDRSetting,
			filetypes );
	if (loadFile(opened))
		setCurrentFile(opened);
}
bool MainGui::loadFile(QString opened) {
if( ! opened.isEmpty() ) {
	QFileInfo qfi(opened);
	// update internal field variable
	RecentDirHDRSetting=qfi.path();
	// if the new dir, the one just chosen by the user, is different from the one stored in the settings, update the settings.
	if (RecentDirHDRSetting != settings.value("RecentDirHDRSetting",QDir::currentPath()).toString()) {
		settings.setValue("RecentDirHDRSetting", RecentDirHDRSetting);
	}
	if (!qfi.isReadable()) {
	    QMessageBox::warning(this,"Aborting...","File is not readable (check existence, permissions,...)",
				 QMessageBox::Ok,QMessageBox::NoButton);
	    return false;
	}
	ImageMDIwindow *nuova;
#ifdef _WIN32
	if (qfi.suffix().toUpper()=="EXR") {
//		readEXRfile(qfi.filePath().toAscii().constData(),"original.pfs");
//		nuova=new ImageMDIwindow( this, QByteArray() );
//		nuova->updateHDR("original.pfs");
	} else if (qfi.suffix().toUpper()=="HDR") {
		readRGBEfile(qfi.filePath().toAscii().constData(),"original.pfs");
		nuova=new ImageMDIwindow(this/*, QByteArray()*/);
		nuova->updateHDR("original.pfs");
	} else if (qfi.suffix().toUpper()=="PFS") {
		QFile::copy(qfi.filePath(),"original.pfs");
		nuova=new ImageMDIwindow(this/*, QByteArray()*/);
		nuova->updateHDR("original.pfs");
	} else {
		QMessageBox::warning(this,"Aborting...","We only support HDR, and PFS files up until now.",
				 QMessageBox::Ok,QMessageBox::NoButton);
		return false;
	}
#else
	if (qfi.suffix().toUpper()=="EXR") {
		QByteArray qba = readEXRfile(qfi.filePath().toAscii().constData());
		nuova=new ImageMDIwindow( this/*, qba*/ );
		nuova->updateHDR(qba);
	} else if (qfi.suffix().toUpper()=="HDR") {
		QByteArray qba =readRGBEfile(qfi.filePath().toAscii().constData());
		nuova=new ImageMDIwindow( this/*, qba*/ );
		nuova->updateHDR(qba);
	} else if (qfi.suffix().toUpper()=="PFS") {
		QFile f(qfi.filePath());
		f.open(QIODevice::ReadOnly);
		nuova=new ImageMDIwindow(this/*, QByteArray()*/);
		nuova->updateHDR(f.readAll());
	} else {
		QMessageBox::warning(this,"Aborting...","We only support EXR, HDR, and PFS files up until now.",
				 QMessageBox::Ok,QMessageBox::NoButton);
		return false;
	}
#endif
	workspace->addWindow(nuova);
	nuova->toolBar->addAction(TonemapAction);
	nuova->setWindowTitle(opened);
	nuova->show();
	return true;
}
return false;
}

void MainGui::fileSaveAs()
{
    QString filetypes;
#ifndef __WIN32
    filetypes += "OpenEXR (*.exr);;";
#endif
    filetypes += "Radiance RGBE (*.hdr *.HDR *.pic *.PIC);;";
    filetypes += "PFS stream (*.pfs *.PFS)";
//     filetypes += "TIFF (*.tif *.tiff *.TIFF *.TIF);;";
//     filetypes += "PFM (*.pfm *.PFM);;";
//     filetypes += "ALL (*)";

	QString fname =QFileDialog::getSaveFileName(
			this,
			"SAVE the HDR to...",
			RecentDirHDRSetting,
			filetypes);

	if(!fname.isEmpty()) {
		QFileInfo qfi(fname);
		// update internal field variable
		RecentDirHDRSetting=qfi.path();
		// if the new dir, the one just chosen by the user, is different from the one stored in the settings, update the settings.
		if (RecentDirHDRSetting != settings.value("RecentDirHDRSetting",QDir::currentPath()).toString()) {
			settings.setValue("RecentDirHDRSetting", RecentDirHDRSetting);
		}
#ifdef _WIN32
		if (qfi.suffix().toUpper()=="EXR") {
//			writeEXRfile  ("original.pfs",qfi.filePath().toAscii().constData());
		} else if (qfi.suffix().toUpper()=="HDR") {
			writeRGBEfile("original.pfs",qfi.filePath().toAscii().constData());
		} else if (qfi.suffix().toUpper()=="PFS") {
			QFile::copy("original.pfs",qfi.filePath());
		} else {
	    		QMessageBox::warning(this,"Aborting...","We only support HDR and PFS files up until now.",
				 QMessageBox::Ok,QMessageBox::NoButton);
		}
#else
		if (qfi.suffix().toUpper()=="EXR") {
			writeEXRfile  (((ImageMDIwindow*)(workspace->activeWindow()))->getHdrPfsStream(),qfi.filePath().toAscii().constData());
		} else if (qfi.suffix().toUpper()=="HDR") {
			writeRGBEfile (((ImageMDIwindow*)(workspace->activeWindow()))->getHdrPfsStream(),qfi.filePath().toAscii().constData());
		} else if (qfi.suffix().toUpper()=="PFS") {
			QFile f(qfi.filePath());
			f.open(QIODevice::WriteOnly);
			f.write(((ImageMDIwindow*)(workspace->activeWindow()))->getHdrPfsStream());
		} else {
	    		QMessageBox::warning(this,"Aborting...","We only support EXR, HDR, and PFS files up until now.",
				 QMessageBox::Ok,QMessageBox::NoButton);
		}
#endif
		setCurrentFile(fname);
	}
}

void MainGui::updatecurrentMDIwindow( QWidget * w )
{
// 	//connect(currentMDIwin, SIGNAL(LumMappingMethodChanged(int)), this, SLOT(MDIwin_changedLumMapping(int)));
	TonemapAction->setEnabled(w!=NULL);
	fileSaveAsAction->setEnabled(w!=NULL);
	rotateccw->setEnabled(w!=NULL);
	rotatecw->setEnabled(w!=NULL);
	menuHDR_Histogram->setEnabled(w!=NULL);
	Low_dynamic_range->setEnabled(w!=NULL);
	Fit_to_dynamic_range->setEnabled(w!=NULL);
	Shrink_dynamic_range->setEnabled(w!=NULL);
	Extend_dynamic_range->setEnabled(w!=NULL);
	Decrease_exposure->setEnabled(w!=NULL);
	Increase_exposure->setEnabled(w!=NULL);
	if (w!=NULL) {
		ImageMDIwindow* current=(ImageMDIwindow*)(workspace->activeWindow());
		if (current->getFittingWin()) {
		normalSizeAct->setEnabled(false);
		zoomInAct->setEnabled(false);
		zoomOutAct->setEnabled(false);
		fitToWindowAct->setEnabled(true);
		} else {
			zoomOutAct->setEnabled(current->getScaleFactor() > 0.222);
			zoomInAct->setEnabled(current->getScaleFactor() < 3.0);
			fitToWindowAct->setEnabled(true);
			normalSizeAct->setEnabled(true);
		}
	} else {
		normalSizeAct->setEnabled(false);
		zoomInAct->setEnabled(false);
		zoomOutAct->setEnabled(false);
		fitToWindowAct->setEnabled(false);
	}
}

void MainGui::tonemap_requested() {
	TMODialog *tmodialog=new TMODialog();
	#ifdef _WIN32
	tmodialog->setOrigBuffer("original.pfs");
	#else
	tmodialog->setOrigBuffer(((ImageMDIwindow*)(workspace->activeWindow()))->getHdrPfsStream());
	#endif
	tmodialog->exec();
	delete tmodialog;
}

void MainGui::rotateccw_requested() {
	dispatchrotate(false);
}

void MainGui::rotatecw_requested() {
	dispatchrotate(true);
}

void MainGui::dispatchrotate( bool clockwise) {
#ifndef _WIN32
	if ((((ImageMDIwindow*)(workspace->activeWindow()))->getHdrPfsStream()).size()==0) {
		QMessageBox::warning(this,"","Problem with original (source) buffer...",
					QMessageBox::Ok, QMessageBox::NoButton);
		return;
	}
#endif
	rotateccw->setEnabled(false);
	rotatecw->setEnabled(false);
	QApplication::setOverrideCursor( QCursor(Qt::WaitCursor) );
#ifdef _WIN32
	//FILE_IN = FILE_OUT IN PFSROTATE!
	rotateFrame("original.pfs",clockwise);
	((ImageMDIwindow*)(workspace->activeWindow()))->updateHDR("original.pfs");
#else
	QByteArray rotated=rotateFrame(((ImageMDIwindow*)(workspace->activeWindow()))->getHdrPfsStream(),clockwise);
	((ImageMDIwindow*)(workspace->activeWindow()))->updateHDR(rotated);
#endif
	QApplication::restoreOverrideCursor();
	rotateccw->setEnabled(true);
	rotatecw->setEnabled(true);
}

void MainGui::current_mdiwindow_decrease_exposure() {
((ImageMDIwindow*)(workspace->activeWindow()))->lumRange->decreaseExposure();
}
void MainGui::current_mdiwindow_extend_exposure() {
((ImageMDIwindow*)(workspace->activeWindow()))->lumRange->extendRange();
}
void MainGui::current_mdiwindow_fit_exposure() {
((ImageMDIwindow*)(workspace->activeWindow()))->lumRange->fitToDynamicRange();
}
void MainGui::current_mdiwindow_increase_exposure() {
((ImageMDIwindow*)(workspace->activeWindow()))->lumRange->increaseExposure();
}
void MainGui::current_mdiwindow_shrink_exposure() {
((ImageMDIwindow*)(workspace->activeWindow()))->lumRange->shrinkRange();
}
void MainGui::current_mdiwindow_ldr_exposure() {
((ImageMDIwindow*)(workspace->activeWindow()))->lumRange->lowDynamicRange();
}
void MainGui::current_mdiwindow_zoomin() {
	ImageMDIwindow* current=(ImageMDIwindow*)(workspace->activeWindow());
	current->zoomIn();
	zoomOutAct->setEnabled(true);
	zoomInAct->setEnabled(current->getScaleFactor() < 3.0);
}
void MainGui::current_mdiwindow_zoomout() {
	ImageMDIwindow* current=(ImageMDIwindow*)(workspace->activeWindow());
	current->zoomOut();
	zoomInAct->setEnabled(true);
	zoomOutAct->setEnabled(current->getScaleFactor() > 0.222);
}
void MainGui::current_mdiwindow_fit_to_win(bool checked) {
	((ImageMDIwindow*)(workspace->activeWindow()))->fitToWindow(checked);
	zoomInAct->setEnabled(!checked);
	zoomOutAct->setEnabled(!checked);
	normalSizeAct->setEnabled(!checked);
}
void MainGui::current_mdiwindow_original_size() {
	((ImageMDIwindow*)(workspace->activeWindow()))->normalSize();
	zoomInAct->setEnabled(true);
	zoomOutAct->setEnabled(true);
}


void MainGui::helpAbout() {
	QDialog *help=new QDialog();
	help->setAttribute(Qt::WA_DeleteOnClose);
	Ui::HelpDialog ui;
	ui.setupUi(help);
	ui.tb->setHtml(helptext);
	help->show();
}

void MainGui::updateRecentFileActions() {
	QStringList files = settings.value("recentFileList").toStringList();
	
	int numRecentFiles = qMin(files.size(), (int)MaxRecentFiles);
	separatorRecentFiles->setVisible(numRecentFiles > 0);
	
	for (int i = 0; i < numRecentFiles; ++i) {
		QString text = tr("&%1 %2").arg(i + 1).arg(strippedName(files[i]));
		recentFileActs[i]->setText(text);
		recentFileActs[i]->setData(files[i]);
		recentFileActs[i]->setVisible(true);
	}
	for (int j = numRecentFiles; j < MaxRecentFiles; ++j)
		recentFileActs[j]->setVisible(false);
}

QString MainGui::strippedName(const QString &fullFileName) {
	return QFileInfo(fullFileName).fileName();
}

void MainGui::openRecentFile() {
	QAction *action = qobject_cast<QAction *>(sender());
	if (action)
		loadFile(action->data().toString());
}

void MainGui::setCurrentFile(const QString &fileName) {
	QStringList files = settings.value("recentFileList").toStringList();
	files.removeAll(fileName);
	files.prepend(fileName);
	while (files.size() > MaxRecentFiles)
		files.removeLast();

	settings.setValue("recentFileList", files);
	updateRecentFileActions();
}

