/**
 * @file pfstmo_pattanaik00.cpp
 * @brief Tone map XYZ channels using Pattanaik00 model
 *
 * Time-Dependent Visual Adaptation for Realistic Image Display
 * S.N. Pattanaik, J. Tumblin, H. Yee, and D.P. Greenberg
 * In Proceedings of ACM SIGGRAPH 2000
 *
 * 
 * This file is a part of PFSTMO package.
 * ---------------------------------------------------------------------- 
 * Copyright (C) 2003,2004 Grzegorz Krawczyk
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ---------------------------------------------------------------------- 
 * 
 * @author Grzegorz Krawczyk, <krawczyk@mpi-sb.mpg.de>
 *
 * $Id: pfstmo_pattanaik00.cpp,v 1.3 2004/12/14 15:12:16 krawczyk Exp $
 */


#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <pfs.h>
#include <tmo_pattanaik00.h>
#include <QFile>

using namespace std;

void multiplyChannels( pfs::Array2D* X, pfs::Array2D* Y, pfs::Array2D* Z, float mult );

#ifdef _WIN32
QByteArray pfstmo_pattanaik00 (const char* input, const char* output, bool _local,float _multiplier, float _Acone, float _Arod, bool autoY ) {
	FILE *inputfp=fopen(input,"rb");
        FILE *fpout=fopen(output,"wb");
#else
QByteArray pfstmo_pattanaik00 (QByteArray inqba, bool _local,float _multiplier, float _Acone, float _Arod, bool autoY ) {
	assert(inqba.size()!=0);
	FILE *inputfp=fmemopen(inqba.data(), (size_t)(inqba.size()), "r");
	//prepare "fake" FILE* where pfs library will write the output hdr to, in pfs stream format.
	char *pfshdr;
	size_t size;
	FILE * fpout=open_memstream (&pfshdr, &size);
#endif
	pfs::DOMIO pfsio;
	//--- default tone mapping parameters;
	bool local = _local;
	float multiplier = _multiplier;
	float Acone = _Acone;
	float Arod  = _Arod;
	
	VisualAdaptationModel* am = new VisualAdaptationModel();
	
	pfs::Frame *frame = pfsio.readFrame( inputfp );
	assert( frame != NULL );
	
	pfs::Channel *X, *Y, *Z;
	frame->getXYZChannels( X, Y, Z );
	
	assert( Y!=NULL || X!=NULL || Z!=NULL);
	
	// adaptation model
	if( multiplier!=1.0f )
		multiplyChannels( X, Y, Z, multiplier );
	
	if( !local ) {
		if( !autoY )
			am->setAdaptation(Acone,Arod);
		else
			am->setAdaptation(Y);
	}
	
	// tone mapping
	int w = Y->getCols();
	int h = Y->getRows();
	pfs::Array2D* R = new pfs::Array2DImpl(w,h);
	pfs::Array2D* G = new pfs::Array2DImpl(w,h);
	pfs::Array2D* B = new pfs::Array2DImpl(w,h);
	
	pfs::transformColorSpace( pfs::CS_XYZ, X, Y, Z, pfs::CS_RGB, R, G, B );
	tmo_pattanaik00( R, G, B, Y, am, local );
	pfs::transformColorSpace( pfs::CS_RGB, R, G, B, pfs::CS_XYZ, X, Y, Z );
	
	delete R;
	delete G;
	delete B;
	delete am; // delete visual adaptation model
	
	pfsio.writeFrame( frame, fpout );
	pfsio.freeFrame( frame );

	fclose(inputfp);
	fclose(fpout);
	QByteArray toreturn;
#ifndef _WIN32
	toreturn=QByteArray::fromRawData(pfshdr,size); //shallow copy
// 	free(pfshdr); //since QByteArray::fromRawData is shallow copy, I cannot free it.
#else
	QFile f(output);
	f.open(QIODevice::ReadOnly);
	toreturn=f.readAll();
#endif
	return toreturn;
}

void multiplyChannels( pfs::Array2D* X, pfs::Array2D* Y, pfs::Array2D* Z, float mult )
{
  int size = Y->getCols() * Y->getRows();

  for( int i=0 ; i<size; i++ )
  {
    (*X)(i) *= mult;
    (*Y)(i) *= mult;
    (*Z)(i) *= mult;
  }
}
