/**
 * @file pfstmo_reinhard04.cpp
 * @brief Tone map XYZ channels using Reinhard04 model
 *
 * Dynamic Range Reduction Inspired by Photoreceptor Physiology.
 * E. Reinhard and K. Devlin.
 * In IEEE Transactions on Visualization and Computer Graphics, 2004.
 *
 * 
 * This file is a part of PFSTMO package.
 * ---------------------------------------------------------------------- 
 * Copyright (C) 2003,2004 Grzegorz Krawczyk
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ---------------------------------------------------------------------- 
 * 
 * @author Grzegorz Krawczyk, <krawczyk@mpi-sb.mpg.de>
 *
 * $Id: pfstmo_reinhard04.cpp,v 1.2 2004/09/22 10:00:30 krawczyk Exp $
 */

// #include <config.h>

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <QByteArray>
#include <QFile>
#include <pfs.h>

using namespace std;

/**
 * @brief: Tone mapping algorithm [Reinhard2004]
 *
 * @param R red channel
 * @param G green channel
 * @param B blue channel
 * @param Y luminance channel
 */
extern "C" void tmo_reinhard04( pfs::Array2D* R, pfs::Array2D* G, pfs::Array2D* B, 
  pfs::Array2D* Y, float f, float w )
{
  float max_lum = (*Y)(0);
  float min_lum = (*Y)(0);
  float world_lum = 0.0;
  int im_width = Y->getCols();
  int im_height = Y->getRows();
  int im_size = im_width * im_height;

  for( int i=1 ; i<im_size ; i++ )
  {
    double lum = (*Y)(i);
    max_lum = (max_lum > lum) ? max_lum : lum;
    min_lum = (min_lum < lum) ? min_lum : lum;
    world_lum += log(2.3e-5+lum);
  }
  world_lum = exp(world_lum/im_size);

  //--- tone map image
  max_lum = log( max_lum );
  min_lum = log( min_lum );

  float k = (max_lum - log(world_lum)) / (max_lum - min_lum);
  float m = 0.3f+0.7f*pow(k,1.4f);
  f = exp(-f);

  float max_col = 0.0f;
  float min_col = 1.0f;

  int x,y;
  for( x=0 ; x<im_width ; x++ )
    for( y=0 ; y<im_height ; y++ )
    {
      float l = (*Y)(x,y);
      float col;
      if( l != 0.0f )
      {
        for( int c=0 ; c<3 ; c++ )
        {
          switch(c)
          {
          case 0: col = (*R)(x,y); break;
          case 1: col = (*G)(x,y); break;
          case 2: col = (*B)(x,y); break;
          };

          if( col!=0.0f)
          {
            float Ia = w*l + (1.0f-w)*col;
            col /= col + pow(f*Ia, m);
          }

          max_col = (col>max_col) ? col : max_col;
          min_col = (col<min_col) ? col : min_col;

          switch(c)
          {
          case 0: (*R)(x,y) = col; break;
          case 1: (*G)(x,y) = col; break;
          case 2: (*B)(x,y) = col; break;
          };
        }
      }
    }

  //--- normalize intensities
  for( x=0 ; x<im_width ; x++ )
    for( y=0 ; y<im_height ; y++ )
    {
      (*R)(x,y) = ((*R)(x,y)-min_col)/(max_col-min_col);
      (*G)(x,y) = ((*G)(x,y)-min_col)/(max_col-min_col);
      (*B)(x,y) = ((*B)(x,y)-min_col)/(max_col-min_col);
    }
}

// extern "C" QByteArray pfstmo_reinhard04(QByteArray inqba, float br, float sat );

#ifdef _WIN32
extern "C" QByteArray pfstmo_reinhard04(const char* input, const char* output, float br, float sat ) {
	FILE *inputfp=fopen(input,"rb");
        FILE *fpout=fopen(output,"wb");
#else
extern "C" QByteArray pfstmo_reinhard04(QByteArray inqba, float br, float sat ) {
	assert(inqba.size()!=0);
	FILE *inputfp=fmemopen(inqba.data(), (size_t)(inqba.size()), "r");
	//prepare "fake" FILE* where pfs library will write the output hdr to, in pfs stream format.
	char *pfshdr;
	size_t size;
	FILE * fpout=open_memstream (&pfshdr, &size);
#endif
	pfs::DOMIO pfsio;
	//--- default tone mapping parameters;
	float brightness = br;
	float saturation = sat;
	
	pfs::Frame *frame = pfsio.readFrame( inputfp );
	assert( frame != NULL );
	
	pfs::Channel *X, *Y, *Z;
	frame->getXYZChannels( X, Y, Z );\
	
	assert( Y!=NULL || X!=NULL || Z!=NULL);
	
	// tone mapping
	int w = Y->getCols();
	int h = Y->getRows();
	pfs::Array2D* R = new pfs::Array2DImpl(w,h);
	pfs::Array2D* G = new pfs::Array2DImpl(w,h);
	pfs::Array2D* B = new pfs::Array2DImpl(w,h);
// 	fprintf(stderr,"prima di r04\n");
	pfs::transformColorSpace( pfs::CS_XYZ, X, Y, Z, pfs::CS_RGB, R, G, B );
	tmo_reinhard04( R, G, B, Y, brightness, saturation );
	pfs::transformColorSpace( pfs::CS_RGB, R, G, B, pfs::CS_XYZ, X, Y, Z );
// 	fprintf(stderr,"dopo di r04\n");
	
	delete R;
	delete G;
	delete B;
	
	pfsio.writeFrame( frame, fpout );
	pfsio.freeFrame( frame );

	fclose(inputfp);
	fclose(fpout);
	QByteArray toreturn;
#ifndef _WIN32
	toreturn=QByteArray::fromRawData(pfshdr,size); //shallow copy
// 	free(pfshdr); //since QByteArray::fromRawData is shallow copy, I cannot free it.
#else
	QFile f(output);
	f.open(QIODevice::ReadOnly);
	toreturn=f.readAll();	
#endif
	return toreturn;
}
